Rank3Tensors := module()
option package;
export 

##########################################################################################
# License: Apache License, Version 2.0
# Copyright [2023] [Joshua James Leiter]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
##########################################################################################

#Update 11/15/2021
#Created a Rank3Tensor Module
#End Update: 11/15/2021

#Update 11/1/2022
#Edited code to remove evalDG where possible.
#End Update 11/1/2022

#Below on 1-7 break a rank 3 tensor into its SO(N) irreducible representations.

#(*1*)
	Rank3FrankA,
#(*2*)
	Rank3TraceA,
#(*3*)
	Rank3FrankB,
#(*4*)
	Rank3TraceB,
#(*5*)
	Rank3FrankC,
#(*6*)
	Rank3TraceC,
#(*7*)	
	Rank3FrankD;
#(*8*)
	
#(*9*)
	
#(*10*)
	
#(*11*)
	
#(*12*)
	
#(*13*)
	
#(*14*)
	
#(*15*)
	
#(*16*)
	
#(*17*)
	
#(*18*)
	
#(*19*)
	
#(*20*)
	
#(*21*)
	
#(*22*)
#(*23*)
#(*24*)
#(*25*)
#(*26*)
#(*27*)
#(*28*)
#(*29*)
#(*30*)
#(*31*)
#(*32*)
#(*33*)
#(*34*)
#(*35*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*1*)

Rank3FrankA:=proc(aTensor,aMetric)
local b1,S1,N,h,Ptensor,TraceA,FrankA;
if DGinfo(aTensor,"TensorIndexType") = ["cov_bas", "cov_bas", "cov_bas"] then
	b1:=YoungTableauBasis([3], [1,2,3], output = "Matrix")[1];
	S1:=YoungSymmetrizer(aTensor,b1);
	N:=DGinfo("FrameBaseDimension");
	h:=InverseMetric(aMetric);
	Ptensor:=ContractIndices(h,S1,[[1,1],[2,2]]);
	TraceA:=SymmetrizeIndices(3/(N+2)*Ptensor &t aMetric,[1,2,3],"Symmetric");
	FrankA:=S1 &minus TraceA;
	return FrankA:
else
   	error "expected 1st argument (metric tensor) to have tensor index type [\"cov_bas\", \"cov_bas\", \"cov_bas\"] %2. Received: %1", DifferentialGeometry:-DGErrorPrint(aTensor)
fi;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*2*)

Rank3TraceA:=proc(aTensor,aMetric)
local b1,S1,N,h,Ptensor,TraceA;
if DGinfo(aTensor,"TensorIndexType") = ["cov_bas", "cov_bas", "cov_bas"] then
	b1:=YoungTableauBasis([3], [1,2,3], output = "Matrix")[1];
	S1:=YoungSymmetrizer(aTensor,b1);
	N:=DGinfo("FrameBaseDimension");
	h:=InverseMetric(aMetric);
	Ptensor:=ContractIndices(h,S1,[[1,1],[2,2]]);
	TraceA:=SymmetrizeIndices(3/(N+2)*Ptensor &t aMetric,[1,2,3],"Symmetric");
	return TraceA:
else
   	error "expected 1st argument (metric tensor) to have tensor index type [\"cov_bas\", \"cov_bas\", \"cov_bas\"] %2. Received: %1", DifferentialGeometry:-DGErrorPrint(aTensor)
fi;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*3*)

Rank3FrankB:=proc(aTensor,aMetric)
local b2,S2,N,h,Ptensor,TraceB,FrankB;
if DGinfo(aTensor,"TensorIndexType") = ["cov_bas", "cov_bas", "cov_bas"] then
	b2:=YoungTableauBasis([2,1], [1,2,3], output = "Matrix")[1];
	S2:=YoungSymmetrizer(aTensor,b2);
	N:=DGinfo("FrameBaseDimension");
	h:=InverseMetric(aMetric);
	Ptensor:=ContractIndices(h,S2,[[1,1],[2,2]]);
	TraceB:=SymmetrizeIndices(-2/(N-1)*Ptensor &t aMetric,[1,3],"SkewSymmetric");
	FrankB:=(S2 &minus TraceB);
	return FrankB:
else
   	error "expected 1st argument (metric tensor) to have tensor index type [\"cov_bas\", \"cov_bas\", \"cov_bas\"] %2. Received: %1", DifferentialGeometry:-DGErrorPrint(aTensor)
fi;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*4*)

Rank3TraceB:=proc(aTensor,aMetric)
local b2,S2,N,h,Ptensor,TraceB;
if DGinfo(aTensor,"TensorIndexType") = ["cov_bas", "cov_bas", "cov_bas"] then
	b2:=YoungTableauBasis([2,1], [1,2,3], output = "Matrix")[1];
	S2:=YoungSymmetrizer(aTensor,b2);
	N:=DGinfo("FrameBaseDimension");
	h:=InverseMetric(aMetric);
	Ptensor:=ContractIndices(h,S2,[[1,1],[2,2]]);
	TraceB:=SymmetrizeIndices(-2/(N-1)*Ptensor &t aMetric,[1,3],"SkewSymmetric");
	return TraceB:
else
   	error "expected 1st argument (metric tensor) to have tensor index type [\"cov_bas\", \"cov_bas\", \"cov_bas\"] %2. Received: %1", DifferentialGeometry:-DGErrorPrint(aTensor)
fi;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*5*)

Rank3FrankC:=proc(aTensor,aMetric)
local b2,S3,N,h,Rtensor,TraceC,FrankC;
if DGinfo(aTensor,"TensorIndexType") = ["cov_bas", "cov_bas", "cov_bas"] then
	b2:=YoungTableauBasis([2,1], [1,2,3], output = "Matrix")[2];
	S3:=YoungSymmetrizer(aTensor,b2);
	N:=DGinfo("FrameBaseDimension");
	h:=InverseMetric(aMetric);
	Rtensor:=ContractIndices(h,S3,[[1,1],[2,3]]);
	TraceC:=SymmetrizeIndices(-2/(N-1)*Rtensor &t aMetric,[1,2],"SkewSymmetric");
	FrankC:=(S3 &minus TraceC):
	return FrankC:
else
   	error "expected 1st argument (metric tensor) to have tensor index type [\"cov_bas\", \"cov_bas\", \"cov_bas\"] %2. Received: %1", DifferentialGeometry:-DGErrorPrint(aTensor)
fi;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*6*)

Rank3TraceC:=proc(aTensor,aMetric)
local b2,S3,N,h,Rtensor,TraceC;
if DGinfo(aTensor,"TensorIndexType") = ["cov_bas", "cov_bas", "cov_bas"] then
	b2:=YoungTableauBasis([2,1], [1,2,3], output = "Matrix")[2];
	S3:=YoungSymmetrizer(aTensor,b2);
	N:=DGinfo("FrameBaseDimension");
	h:=InverseMetric(aMetric);
	Rtensor:=ContractIndices(h,S3,[[1,1],[2,3]]);
	TraceC:=SymmetrizeIndices(-2/(N-1)*Rtensor &t aMetric,[1,2],"SkewSymmetric");
	return TraceC:
else
   	error "expected 1st argument (metric tensor) to have tensor index type [\"cov_bas\", \"cov_bas\", \"cov_bas\"] %2. Received: %1", DifferentialGeometry:-DGErrorPrint(aTensor)
fi;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*7*)

Rank3FrankD:=proc(aTensor)
local b3,S4;
if DGinfo(aTensor,"TensorIndexType") = ["cov_bas", "cov_bas", "cov_bas"] then
	b3:=YoungTableauBasis([1,1,1], [1,2,3], output = "Matrix")[1];
	S4:=YoungSymmetrizer(aTensor,b3);
	return S4:
else
   	error "expected 1st argument (metric tensor) to have tensor index type [\"cov_bas\", \"cov_bas\", \"cov_bas\"] %2. Received: %1", DifferentialGeometry:-DGErrorPrint(aTensor)
fi;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*8*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*9*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*10*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*11*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*12*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*13*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*14*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*15*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*16*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*17*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*18*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*19*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*20*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*21*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*22*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*23*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*24*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*25*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*26*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*27*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*28*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*29*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*30*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*31*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*32*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*33*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*34*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*35*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*36*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*37*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*38*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*39*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*40*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*41*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*42*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*43*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*44*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*45*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*46*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*47*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*48*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*49*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*50*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*51*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*52*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*53*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*54*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*55*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*56*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*57*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*58*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*59*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*60*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*61*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*62*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*63*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*64*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*65*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*66*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*67*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*68*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*69*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*70*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*71*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*72*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*73*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*74*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*75*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*76*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*77*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*78*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*79*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*80*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*81*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*82*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*83*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*84*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*85*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*86*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*87*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*88*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*89*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*90*)



#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111

end module: