SegreInvariants := module()
option package;
export 

##########################################################################################
# License: Apache License, Version 2.0
# Copyright [2023] [Joshua James Leiter]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
##########################################################################################

#Update 10/13/2022
#Created a SegreInvariants Module for the calculation of these in regards to the Segre Classification.
#End Update: 10/12/2022

#Update 11/2/2022
#Added PetrovPsiCovariant for calculation in the full Curvature Classification Algorithm
#End Update: 11/2/2022

#Update 2/22/2023
#Added Petrov Covariants and Invariants for further classification analysis.
#These are number (16)-(20)
#End Update: 2/22/2023

#Update 2/27/2023
#Updated to include quotes in some of the LHS pieces in the lists
#End Update 2/27/2023

#Below are the Segre invariants from Zakhary and Carminati. These are in Commands 1-14

#(*1*)
	SegrePhiCovariant,
#(*2*)
	SegreChiCovariant,
#(*3*)
	SegreECovariant,
#(*4*)
	SegreChiTildeCovariant,
#(*5*)
	SegreQTildeCovariant,
#(*6*)
	SegreRTildeCovariant,
#(*7*)
	SegreI6Invariant,
#(*8*)
	SegreI7Invariant,
#(*9*)
	SegreI8Invariant,
#(*10*)
	SegreIpInvariant,
#(*11*)
	SegreJpInvariant,
#(*12*)
	SegreDeltaInvariant,
#(*13*)
	SegreHInvariant,
#(*14*)
	SegreKaCovariant,
#(*15*)
	PetrovPsiCovariant,
#(*16*)
	PetrovQCovariant,
#(*17*)
	PetrovRCovariant,
#(*18*)
	PetrovIpInvariant,
#(*19*)
	PetrovJpInvariant,
#(*20*)
	PetrovDeltaInvariant;
#(*21*)
	
#(*22*)
	
#(*23*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*1*)

SegrePhiCovariant:=proc(aPhiSpinor)
local Dz1,Dz2,Dw1,Dw2,Phi00,Phi01,Phi02,Phi10,Phi11,Phi12,Phi20,Phi21,Phi22,myList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

Dz1:=DGinfo("FrameFiberVectors")[1];
Dz2:=DGinfo("FrameFiberVectors")[2];
Dw1:=DGinfo("FrameFiberVectors")[3];
Dw2:=DGinfo("FrameFiberVectors")[4];

#Build the Phi Scalars:
Phi00:=simplify(Hook([Dz1,Dz1,Dw1,Dw1],aPhiSpinor));
Phi01:=simplify(Hook([Dz1,Dz1,Dw1,Dw2],aPhiSpinor));
Phi02:=simplify(Hook([Dz1,Dz1,Dw2,Dw2],aPhiSpinor));
Phi10:=simplify(Hook([Dz1,Dz2,Dw1,Dw1],aPhiSpinor));
Phi11:=simplify(Hook([Dz1,Dz2,Dw1,Dw2],aPhiSpinor));
Phi12:=simplify(Hook([Dz1,Dz2,Dw2,Dw2],aPhiSpinor));
Phi20:=simplify(Hook([Dz2,Dz2,Dw1,Dw1],aPhiSpinor));
Phi21:=simplify(Hook([Dz2,Dz2,Dw1,Dw2],aPhiSpinor));
Phi22:=simplify(Hook([Dz2,Dz2,Dw2,Dw2],aPhiSpinor));

myList:=[
"Phi00"=Phi00,
"Phi01"=Phi01,
"Phi02"=Phi02,
"Phi10"=Phi10,
"Phi11"=Phi11,
"Phi12"=Phi12,
"Phi20"=Phi20,
"Phi21"=Phi21,
"Phi22"=Phi22];

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*2*)

SegreChiCovariant:=proc(aPhiSpinor)
local myList1,myEvalList,Phi00,Phi01,Phi02,Phi10,Phi11,Phi12,Phi20,Phi21,Phi22,myList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=SegrePhiCovariant(aPhiSpinor);

myEvalList:=[
Phi00=rhs(myList1[1]),
Phi01=rhs(myList1[2]),
Phi02=rhs(myList1[3]),
Phi10=rhs(myList1[4]),
Phi11=rhs(myList1[5]),
Phi12=rhs(myList1[6]),
Phi20=rhs(myList1[7]),
Phi21=rhs(myList1[8]),
Phi22=rhs(myList1[9])];

#Build the Chi Scalars from the Phi Scalars:
myList:=simplify(eval([
"chi0"=2*Phi00*Phi02-2*Phi01^2,
"chi1"=Phi00*Phi12-2*Phi01*Phi11+Phi02*Phi10,
"chi2"=(Phi00*Phi22-2*Phi01*Phi21+Phi02*Phi20+4*Phi12*Phi10-4*Phi11^2)/3,
"chi3"=Phi10*Phi22-2*Phi11*Phi21+Phi12*Phi20,
"chi4"=2*Phi20*Phi22-2*Phi21^2],myEvalList));

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*3*)

SegreECovariant:=proc(aPhiSpinor)
local myList1,myEvalList,Phi00,Phi01,Phi02,Phi10,Phi11,Phi12,Phi20,Phi21,Phi22,myList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=SegrePhiCovariant(aPhiSpinor);

myEvalList:=[
Phi00=rhs(myList1[1]),
Phi01=rhs(myList1[2]),
Phi02=rhs(myList1[3]),
Phi10=rhs(myList1[4]),
Phi11=rhs(myList1[5]),
Phi12=rhs(myList1[6]),
Phi20=rhs(myList1[7]),
Phi21=rhs(myList1[8]),
Phi22=rhs(myList1[9])];

#Build the E Scalars:
myList:=simplify(eval([
"E00" = 4*Phi00*Phi11-4*Phi01*Phi10,
"E01" = 2*Phi00*Phi12-2*Phi02*Phi10,
"E02" = 4*Phi01*Phi12-4*Phi02*Phi11,
"E10" = 2*Phi00*Phi21-2*Phi01*Phi20,
"E11" = Phi00*Phi22-Phi02*Phi20,
"E12" = 2*Phi01*Phi22-2*Phi02*Phi21,
"E20" = 4*Phi10*Phi21-4*Phi11*Phi20,
"E21" = 2*Phi10*Phi22-2*Phi12*Phi20,
"E22" = 4*Phi11*Phi22-4*Phi12*Phi21],myEvalList));

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*4*)

SegreChiTildeCovariant:=proc(aPhiSpinor)
local myList1,myEvalList,E00,E01,E02,E10,E11,E12,E20,E21,E22,myList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=SegreECovariant(aPhiSpinor);

myEvalList:=[
E00=rhs(myList1[1]),
E01=rhs(myList1[2]),
E02=rhs(myList1[3]),
E10=rhs(myList1[4]),
E11=rhs(myList1[5]),
E12=rhs(myList1[6]),
E20=rhs(myList1[7]),
E21=rhs(myList1[8]),
E22=rhs(myList1[9])];

#Build the ChiTilde Scalars from the E Scalars:
myList:=simplify(eval([
"chiTilde0"=2*E00*E02-2*E01^2,
"chiTilde1"=E00*E12-2*E01*E11+E02*E10,
"chiTilde2"=(1/3)*E00*E22-(2/3)*E01*E21+(1/3)*E02*E20+(4/3)*E12*E10-(4/3)*E11^2,
"chiTilde3"=E10*E22-2*E11*E21+E12*E20,
"chiTilde4"=2*E20*E22-2*E21^2],myEvalList));

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*5*)

SegreQTildeCovariant:=proc(aPhiSpinor)
local myList1,myEvalList,chi0,chi1,chi2,chi3,chi4,myList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=SegreChiCovariant(aPhiSpinor);

myEvalList:=[
chi0=rhs(myList1[1]),
chi1=rhs(myList1[2]),
chi2=rhs(myList1[3]),
chi3=rhs(myList1[4]),
chi4=rhs(myList1[5])];

#Build the Q Scalars from the chi Scalars:
myList:=simplify(eval([
"QTilde0"=2*chi0*chi2-2*chi1^2,
"QTilde1"=chi0*chi3-chi1*chi2,
"QTilde2"=(1/3)*chi0*chi4+(2/3)*chi1*chi3-chi2^2,
"QTilde3"=chi1*chi4-chi2*chi3,
"QTilde4"=2*chi2*chi4-2*chi3^2],myEvalList));

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*6*)

SegreRTildeCovariant:=proc(aPhiSpinor)
local myList1,myList2,myEvalList1,myEvalList2,chi0,chi1,chi2,chi3,chi4,
QTilde0,QTilde1,QTilde2,QTilde3,QTilde4,myEvalList,myList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=SegreChiCovariant(aPhiSpinor);
myList2:=SegreQTildeCovariant(aPhiSpinor);

myEvalList1:=[
chi0=rhs(myList1[1]),
chi1=rhs(myList1[2]),
chi2=rhs(myList1[3]),
chi3=rhs(myList1[4]),
chi4=rhs(myList1[5])];

myEvalList2:=[
QTilde0=rhs(myList2[1]),
QTilde1=rhs(myList2[2]),
QTilde2=rhs(myList2[3]),
QTilde3=rhs(myList2[4]),
QTilde4=rhs(myList2[5])];

myEvalList:=[op(myEvalList1),op(myEvalList2)];

#Build the R Scalars from the QTilde & chi Scalars:
myList:=simplify(eval([
"Rtilde0"=QTilde0*chi1-QTilde1*chi0,
"Rtilde1"=(QTilde0*chi2-QTilde2*chi0)/2,
"Rtilde2"=QTilde1*chi2-QTilde2*chi1,
"Rtilde3"=(QTilde1*chi3-QTilde3*chi1)/2,
"Rtilde4"=QTilde2*chi3-QTilde3*chi2,
"Rtilde5"=(QTilde2*chi4-QTilde4*chi2)/2,
"Rtilde6"=QTilde3*chi4-QTilde4*chi3],myEvalList));

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*7*)

SegreI6Invariant:=proc(aPhiSpinor)
local myList1,myEvalList,Phi00,Phi01,Phi02,Phi10,Phi11,Phi12,Phi20,Phi21,Phi22,myList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=SegrePhiCovariant(aPhiSpinor);

myEvalList:=[
Phi00=rhs(myList1[1]),
Phi01=rhs(myList1[2]),
Phi02=rhs(myList1[3]),
Phi10=rhs(myList1[4]),
Phi11=rhs(myList1[5]),
Phi12=rhs(myList1[6]),
Phi20=rhs(myList1[7]),
Phi21=rhs(myList1[8]),
Phi22=rhs(myList1[9])];

#Build the I6 (r1) invariant from the Phi Scalars:
myList:=simplify(eval([
"I6"=(2/3)*Phi00*Phi22-(4/3)*Phi01*Phi21+(2/3)*Phi02*Phi20-(4/3)*Phi12*Phi10+(4/3)*Phi11^2],myEvalList));

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*8*)

SegreI7Invariant:=proc(aPhiSpinor)
local myList1,myEvalList,Phi00,Phi01,Phi02,Phi10,Phi11,Phi12,Phi20,Phi21,Phi22,myList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=SegrePhiCovariant(aPhiSpinor);

myEvalList:=[
Phi00=rhs(myList1[1]),
Phi01=rhs(myList1[2]),
Phi02=rhs(myList1[3]),
Phi10=rhs(myList1[4]),
Phi11=rhs(myList1[5]),
Phi12=rhs(myList1[6]),
Phi20=rhs(myList1[7]),
Phi21=rhs(myList1[8]),
Phi22=rhs(myList1[9])];

#Build the Chi Scalars from the Phi Scalars:
myList:=simplify(eval([
"I7"=2*Phi00*Phi11*Phi22-2*Phi00*Phi12*Phi21-2*Phi01*Phi10*Phi22+2*Phi01*Phi12*Phi20+2*Phi02*Phi10*Phi21-2*Phi02*Phi11*Phi20],myEvalList));

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*9*)

SegreI8Invariant:=proc(aPhiSpinor)
local myList1,myEvalList,Phi00,Phi01,Phi02,Phi10,Phi11,Phi12,Phi20,Phi21,Phi22,myList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=SegrePhiCovariant(aPhiSpinor);

myEvalList:=[
Phi00=rhs(myList1[1]),
Phi01=rhs(myList1[2]),
Phi02=rhs(myList1[3]),
Phi10=rhs(myList1[4]),
Phi11=rhs(myList1[5]),
Phi12=rhs(myList1[6]),
Phi20=rhs(myList1[7]),
Phi21=rhs(myList1[8]),
Phi22=rhs(myList1[9])];

#Build the Chi Scalars from the Phi Scalars:
myList:=simplify(eval([
"I8"=(2/3)*Phi00^2*Phi22^2-(8/3)*Phi00*Phi01*Phi21*Phi22+4*Phi00*Phi02*Phi20*Phi22-(8/3)*Phi00*Phi02*Phi21^2-(8/3)*Phi00*Phi10*Phi12*Phi22+(16/3)*Phi00*Phi11*Phi12*Phi21-(8/3)*Phi00*Phi12^2*Phi20-(8/3)*Phi01^2*Phi20*Phi22+(16/3)*Phi01^2*Phi21^2-(8/3)*Phi01*Phi02*Phi20*Phi21+(16/3)*Phi01*Phi10*Phi11*Phi22-(32/3)*Phi01*Phi11^2*Phi21+(16/3)*Phi01*Phi11*Phi12*Phi20+(2/3)*Phi02^2*Phi20^2-(8/3)*Phi02*Phi10^2*Phi22+(16/3)*Phi02*Phi10*Phi11*Phi21-(8/3)*Phi02*Phi10*Phi12*Phi20+(16/3)*Phi10^2*Phi12^2-(32/3)*Phi10*Phi11^2*Phi12+(16/3)*Phi11^4],myEvalList));

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*10*)

SegreIpInvariant:=proc(aPhiSpinor)
local myList1,myEvalList,chi0,chi1,chi2,chi3,chi4,myList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=SegreChiCovariant(aPhiSpinor);

myEvalList:=[
chi0=rhs(myList1[1]),
chi1=rhs(myList1[2]),
chi2=rhs(myList1[3]),
chi3=rhs(myList1[4]),
chi4=rhs(myList1[5])];

#Build the ChiTilde Scalars from the E Scalars:
myList:=simplify(eval([
"IP"=(1/3)*chi0*chi4-(4/3)*chi1*chi3+chi2^2],myEvalList));

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*11*)

SegreJpInvariant:=proc(aPhiSpinor)
local myList1,myEvalList,chi0,chi1,chi2,chi3,chi4,myList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=SegreChiCovariant(aPhiSpinor);

myEvalList:=[
chi0=rhs(myList1[1]),
chi1=rhs(myList1[2]),
chi2=rhs(myList1[3]),
chi3=rhs(myList1[4]),
chi4=rhs(myList1[5])];

#Build the ChiTilde Scalars from the E Scalars:
myList:=simplify(eval([
"JP"=chi0*chi2*chi4-chi0*chi3^2-chi1^2*chi4+2*chi1*chi2*chi3-chi2^3],myEvalList));

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*12*)

SegreDeltaInvariant:=proc(aPhiSpinor)
local myIP,myJP,myList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myIP:=rhs(SegreIpInvariant(aPhiSpinor)[1]);
myJP:=rhs(SegreJpInvariant(aPhiSpinor)[1]);

#Build the Delta Scalar (Discriminant) from the Jp and Ip invariants.
myList:=["Delta"=simplify(myJP^2-myIP^3)];

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*13*)

SegreHInvariant:=proc(aPhiSpinor)
local myList1,myList2,myEvalList1,Phi00,Phi01,Phi02,Phi10,Phi11,Phi12,Phi20,Phi21,Phi22,
chi0,chi1,chi2,chi3,chi4,myEvalList2,myEvalList,myList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=SegrePhiCovariant(aPhiSpinor);

myEvalList1:=[
Phi00=rhs(myList1[1]),
Phi01=rhs(myList1[2]),
Phi02=rhs(myList1[3]),
Phi10=rhs(myList1[4]),
Phi11=rhs(myList1[5]),
Phi12=rhs(myList1[6]),
Phi20=rhs(myList1[7]),
Phi21=rhs(myList1[8]),
Phi22=rhs(myList1[9])];

myList2:=SegreChiCovariant(aPhiSpinor);

myEvalList2:=[
chi0=rhs(myList2[1]),
chi1=rhs(myList2[2]),
chi2=rhs(myList2[3]),
chi3=rhs(myList2[4]),
chi4=rhs(myList2[5])];

myEvalList:=[op(myEvalList1),op(myEvalList2)];

#Build the Chi Scalars from the Phi Scalars:
myList:=simplify(eval([
"H"=3*chi2^3+(1/3*(-12*Phi10*Phi12+12*Phi11^2))*chi2^2+(1/3*(-chi0*chi4-14*chi1*chi3))*chi2+(1/3*(3*chi3^2-4*chi4*(Phi10*Phi12-Phi11^2)))*chi0+chi1^2*chi4+(16/3)*chi3*(Phi10*Phi12-Phi11^2)*chi1],myEvalList));

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*14*)

SegreKaCovariant:=proc(aPhiSpinor)
local myList1,myList2,Phi00,Phi01,Phi02,Phi10,Phi11,Phi12,Phi20,Phi21,Phi22,
E00,E01,E02,E10,E11,E12,E20,E21,E22,
myEvalList1,myEvalList2,my_Ip,my_Jp,my_I7,myH,myEvalList,myList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=SegrePhiCovariant(aPhiSpinor);

myList2:=SegreECovariant(aPhiSpinor);

myEvalList1:=[
Phi00=rhs(myList1[1]),
Phi01=rhs(myList1[2]),
Phi02=rhs(myList1[3]),
Phi10=rhs(myList1[4]),
Phi11=rhs(myList1[5]),
Phi12=rhs(myList1[6]),
Phi20=rhs(myList1[7]),
Phi21=rhs(myList1[8]),
Phi22=rhs(myList1[9])];

myEvalList2:=[
E00=rhs(myList2[1]),
E01=rhs(myList2[2]),
E02=rhs(myList2[3]),
E10=rhs(myList2[4]),
E11=rhs(myList2[5]),
E12=rhs(myList2[6]),
E20=rhs(myList2[7]),
E21=rhs(myList2[8]),
E22=rhs(myList2[9])];

my_Jp:=rhs(SegreJpInvariant(aPhiSpinor)[1]);
my_I7:=rhs(SegreI7Invariant(aPhiSpinor)[1]);
myH:=rhs(SegreHInvariant(aPhiSpinor)[1]);
my_Ip:=rhs(SegreIpInvariant(aPhiSpinor)[1]);

myEvalList:=[op(myEvalList1),op(myEvalList2)];

#Build the ChiTilde Scalars from the E Scalars:
myList:=simplify(eval([
"k0"=my_Jp*(E00-2*my_I7*my_Ip*Phi00/myH),
"k1"=my_Jp*(E11-2*my_I7*my_Ip*Phi11/myH),
"k2"=my_Jp*(E22-2*my_I7*my_Ip*Phi22/myH)],myEvalList));

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*15*)

PetrovPsiCovariant:=proc(aPsiSpinor)
local symmetryCheck1,myFiberBasis,Dz1,Dz2,myPsi0,myPsi1,myPsi2,myPsi3,myPsi4,myList;

#Check if symmetric:
symmetryCheck1:=SymmetrizeIndices(aPsiSpinor,[1,2,3,4],"Symmetric");

if DGequal(symmetryCheck1,aPsiSpinor)<>true then
error "invalid spinor type, expecting type symmetric on indices [1,2,3,4], Recieved:", DGErrorPrint(aPsiSpinor)
fi;

myFiberBasis:=DGinfo("FrameFiberVectors");

Dz1:=myFiberBasis[1];
Dz2:=myFiberBasis[2];

myPsi0:=Hook([Dz1,Dz1,Dz1,Dz1],aPsiSpinor);
myPsi1:=Hook([Dz1,Dz1,Dz1,Dz2],aPsiSpinor);
myPsi2:=Hook([Dz1,Dz1,Dz2,Dz2],aPsiSpinor);
myPsi3:=Hook([Dz1,Dz2,Dz2,Dz2],aPsiSpinor);
myPsi4:=Hook([Dz2,Dz2,Dz2,Dz2],aPsiSpinor);

myList:=["Psi0"=myPsi0,"Psi1"=myPsi1,"Psi2"=myPsi2,"Psi3"=myPsi3,"Psi4"=myPsi4];
return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*16*)

PetrovQCovariant:=proc(aPsiSpinor)
local symmetryCheck1,myList1,Psi0,Psi1,Psi2,Psi3,Psi4,myEvalList,myList;

#Check if symmetric:
symmetryCheck1:=SymmetrizeIndices(aPsiSpinor,[1,2,3,4],"Symmetric");

if DGequal(symmetryCheck1,aPsiSpinor)<>true then
error "invalid spinor type, expecting type symmetric on indices [1,2,3,4], Recieved:", DGErrorPrint(aPsiSpinor)
fi;

myList1:=PetrovPsiCovariant(aPsiSpinor);

myEvalList:=[
Psi0=rhs(myList1[1]),
Psi1=rhs(myList1[2]),
Psi2=rhs(myList1[3]),
Psi3=rhs(myList1[4]),
Psi4=rhs(myList1[5])];

#Build the Q Scalars from the Psi Scalars:
myList:=simplify(eval([
"Q0"=2*Psi0*Psi2-2*Psi1^2,
"Q1"=Psi0*Psi3-Psi1*Psi2,
"Q2"=1/3*Psi0*Psi4+2/3*Psi1*Psi3-Psi2^2,
"Q3"=Psi1*Psi4-Psi2*Psi3,
"Q4"=2*Psi2*Psi4-2*(Psi3)^2],myEvalList));

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*17*)

PetrovRCovariant:=proc(aPsiSpinor)
local symmetryCheck1,myList1,myList2,Psi0,Psi1,Psi2,Psi3,Psi4,Q0,Q1,Q2,Q3,Q4,myEvalList1,myEvalList2,myEvalList,myList;

#Check if symmetric:
symmetryCheck1:=SymmetrizeIndices(aPsiSpinor,[1,2,3,4],"Symmetric");

if DGequal(symmetryCheck1,aPsiSpinor)<>true then
error "invalid spinor type, expecting type symmetric on indices [1,2,3,4], Recieved:", DGErrorPrint(aPsiSpinor)
fi;

myList1:=PetrovPsiCovariant(aPsiSpinor);
myList2:=PetrovQCovariant(aPsiSpinor);

myEvalList1:=[
Psi0=rhs(myList1[1]),
Psi1=rhs(myList1[2]),
Psi2=rhs(myList1[3]),
Psi3=rhs(myList1[4]),
Psi4=rhs(myList1[5])];

myEvalList2:=[
Q0=rhs(myList2[1]),
Q1=rhs(myList2[2]),
Q2=rhs(myList2[3]),
Q3=rhs(myList2[4]),
Q4=rhs(myList2[5])];

myEvalList:=[op(myEvalList1),op(myEvalList2)];

#Build the R Scalars from the Psi and Q Scalars:
myList:=simplify(eval([
"R0"=Q0*Psi1-Q1*Psi0,
"R1"=1/2*(Q0*Psi2-Q2*Psi0),
"R2"=Q1*Psi2-Q2*Psi1,
"R3"=1/2*(Q1*Psi3-Q3*Psi1),
"R4"=Q2*Psi3-Q3*Psi2,
"R5"=1/2*(Q2*Psi4-Q4*Psi2),
"R6"=Q3*Psi4-Q4*Psi3],myEvalList));

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*18*)

PetrovIpInvariant:=proc(aPsiSpinor)
local symmetryCheck1,myList1,Psi0,Psi1,Psi2,Psi3,Psi4,myEvalList,myList;

#Check if symmetric:
symmetryCheck1:=SymmetrizeIndices(aPsiSpinor,[1,2,3,4],"Symmetric");

if DGequal(symmetryCheck1,aPsiSpinor)<>true then
error "invalid spinor type, expecting type symmetric on indices [1,2,3,4], Recieved:", DGErrorPrint(aPsiSpinor)
fi;

myList1:=PetrovPsiCovariant(aPsiSpinor);

myEvalList:=[
Psi0=rhs(myList1[1]),
Psi1=rhs(myList1[2]),
Psi2=rhs(myList1[3]),
Psi3=rhs(myList1[4]),
Psi4=rhs(myList1[5])];

#Build the Ip Invariant from the Psi Scalars:
myList:=simplify(eval([
"Ip"=1/3*Psi0*Psi4-4/3*Psi1*Psi3+(Psi2)^2],myEvalList));

return myList;
end proc:
#1111111111111111111111111111111111111111111111111111111111111111111111
#(*19*)

PetrovJpInvariant:=proc(aPsiSpinor)
local symmetryCheck1,myList1,Psi0,Psi1,Psi2,Psi3,Psi4,myEvalList,myList;

#Check if symmetric:
symmetryCheck1:=SymmetrizeIndices(aPsiSpinor,[1,2,3,4],"Symmetric");

if DGequal(symmetryCheck1,aPsiSpinor)<>true then
error "invalid spinor type, expecting type symmetric on indices [1,2,3,4], Recieved:", DGErrorPrint(aPsiSpinor)
fi;

myList1:=PetrovPsiCovariant(aPsiSpinor);

myEvalList:=[
Psi0=rhs(myList1[1]),
Psi1=rhs(myList1[2]),
Psi2=rhs(myList1[3]),
Psi3=rhs(myList1[4]),
Psi4=rhs(myList1[5])];

#Build the Q Scalars from the Psi Scalars:
myList:=simplify(eval([
"Jp"=Psi0*Psi2*Psi4+2*Psi1*Psi2*Psi3-Psi0*Psi3^2-Psi1^2*Psi4-Psi2^3],myEvalList));

return myList;
end proc:
#1111111111111111111111111111111111111111111111111111111111111111111111
#(*20*)

PetrovDeltaInvariant:=proc(aPsiSpinor)
local symmetryCheck1,myList1,myList2,myEvalList1,myEvalList2,myEvalList,myList;

#Check if symmetric:
symmetryCheck1:=SymmetrizeIndices(aPsiSpinor,[1,2,3,4],"Symmetric");

if DGequal(symmetryCheck1,aPsiSpinor)<>true then
error "invalid spinor type, expecting type symmetric on indices [1,2,3,4], Recieved:", DGErrorPrint(aPsiSpinor)
fi;

myList1:=PetrovIpInvariant(aPsiSpinor);
myList2:=PetrovJpInvariant(aPsiSpinor);

myEvalList1:=[
Ip=rhs(myList1[1])];
myEvalList2:=[
Jp=rhs(myList2[1])];

myEvalList:=[op(myEvalList1),op(myEvalList2)];

#Build the Delta Invariant from the Ip and Jp Invariantss:
myList:=simplify(eval([
"Delta"=Ip^3-Jp^2],myEvalList));

return myList;
end proc:
#1111111111111111111111111111111111111111111111111111111111111111111111
#(*21*)


#1111111111111111111111111111111111111111111111111111111111111111111111
#(*22*)


#1111111111111111111111111111111111111111111111111111111111111111111111
#(*23*)


#1111111111111111111111111111111111111111111111111111111111111111111111
#(*24*)


#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111

end module: